@@ -3,7 +3,7 @@ import datetime, math
 
 import json
 import re
-import sys
+import sys, io, base64
 import requests
 import pprint
 from typing import Optional, Sequence, Union, List
@@ -236,22 +236,44 @@ class ByLocation(GridPlacement):
 
 
 class BulkOperationResult:
+    """
+    Encapsulates the return value from any bulk operation
+    """
+
     def __init__(self, json):
         self.data = json
 
     def is_ok(self):
         return self.data["status"] == "COMPLETED"
 
+    def results(self):
+        return self.data["results"]
+
+    def success_results(self):
+        """
+        Returns results as list of dicts{record: error:} where result was successful
+        """
+        return list(filter(lambda x: x["record"] is not None, self.results()))
+
+    def error_results(self):
+        """
+        Returns results as list of dicts{record: error:} where  error field is not None
+        """
+        return list(filter(lambda x: x["error"] is not None, self.results()))
+
     def is_failed(self):
         return not self.is_ok()
 
     def __str__(self):
         return f"Succeeded: {self.is_ok()}: Result JSON: {self.data}"
 
+    def __repr__(self):
+        return f"Succeeded: {self.is_ok()}: Result JSON: {self.data!r}"
+
 
 class Container:
     """
-    Base class of all Container types
+    Base class of all Container types (representing Container data obtained from RSpace).
     """
 
     @classmethod
@@ -281,6 +303,8 @@ class Container:
             return ListContainer(container)
         elif container["cType"] == "WORKBENCH":
             return Workbench(container)
+        elif container["cType"] == "IMAGE":
+            return ImageContainer(container)
         else:
             raise ValueError(f"unsupported container type {container['cType']}")
 
@@ -310,6 +334,9 @@ class Container:
     def is_workbench(self) -> bool:
         return False
 
+    def is_image(self) -> bool:
+        return False
+
     def accept_subsamples(self) -> bool:
         return self.data["canStoreSamples"] == True
 
@@ -343,6 +370,44 @@ class ListContainer(Container):
 storesSubsamples={self.accept_subsamples}"
 
 
+class ImageContainer(Container):
+    """
+    Wrapper around dict of ImageContainer JSON
+    """
+
+    def __init__(self, image_container: dict):
+        super().__init__(image_container)
+        self._validate_type(image_container, "IMAGE")
+
+    def is_image(self) -> bool:
+        return True
+
+    def capacity(self) -> int:
+        """
+        Returns number of locations defined.
+        """
+        return len(self.data["locations"])
+
+    def free_locations(self) -> int:
+        """
+        Returns
+        -------
+        int
+            Number of locations with no content.
+
+        """
+        return len(list(filter(lambda x: x["content"] is None, self.data["locations"])))
+
+    def used_locations(self) -> int:
+        """
+        Returns
+        -------
+        int
+            Number of locations with content.
+        """
+        return self.capacity() - self.free_locations()
+
+
 class Workbench(Container):
     """
     A specialised Container holding currently active samples and containers.
@@ -486,6 +551,14 @@ class Id:
         "IT": "sampleTemplates",
     }
 
+    @staticmethod
+    def is_valid_id(arg):
+        try:
+            Id(arg)
+        except ValueError:
+            return False
+        return True
+
     def __init__(self, value: Union[int, str, dict, Container, Workbench, Sample]):
 
         if isinstance(value, str):
@@ -518,7 +591,7 @@ class Id:
             self.id = value
         else:
             raise TypeError(
-                f"Could not interpet {value} as an identifiable Inventory item."
+                f"Could not interpet {value!r} as an identifiable Inventory item."
             )
 
     def __repr__(self):
@@ -566,6 +639,12 @@ class Id:
     def is_subsample(self, maybe: bool = False) -> bool:
         return self._check("SS", maybe)
 
+    def is_bench(self, maybe: bool = False) -> bool:
+        return self._check("BE", maybe)
+
+    def is_sample(self, maybe: bool = False) -> bool:
+        return self._check("SA", maybe)
+
     def is_movable(self, maybe: bool = False) -> bool:
         return self.is_subsample(maybe) or self.is_container(maybe)
 
@@ -668,6 +747,301 @@ class ExtraField:
 {self.data['content']!r})"
 
 
+class ItemPost:
+    """
+    Help define core properties of an Inventory item
+    """
+
+    def __init__(
+        self,
+        name: str,
+        tags: Optional[str] = None,
+        description: Optional[str] = None,
+        extra_fields: Optional[Sequence] = [],
+    ):
+        self.data = {}
+        self.data["name"] = name
+        if tags is not None:
+            self.data["tags"] = tags
+        if description is not None:
+            self.data["description"] = description
+        if extra_fields is not None:
+            self.data["extraFields"] = [ef.data for ef in extra_fields]
+
+
+class SamplePost(ItemPost):
+    """
+    Help define sample data structures to create or modify samples
+    """
+
+    def __init__(
+        self,
+        name: str,
+        tags: Optional[str] = None,
+        description: Optional[str] = None,
+        extra_fields: Optional[Sequence] = [],
+        sample_template_id=None,
+        fields=None,
+        storage_temperature_min: StorageTemperature = None,
+        storage_temperature_max: StorageTemperature = None,
+        expiry_date: datetime.datetime = None,
+        subsample_count: int = None,
+        total_quantity: Quantity = None,
+        attachments=None,
+    ):
+        super().__init__(name, tags, description, extra_fields)
+        ## converts arguments into JSON POST syntax
+        self.data["type"] = "SAMPLE"
+        if storage_temperature_min is not None:
+            self.data["storageTempMin"] = storage_temperature_min._toDict()
+        if storage_temperature_max is not None:
+            self.data["storageTempMax"] = storage_temperature_max._toDict()
+        if expiry_date is not None:
+            self.data["expiryDate"] = expiry_date.isoformat()
+        if subsample_count is not None:
+            self.data["newSampleSubSamplesCount"] = subsample_count
+        if total_quantity is not None:
+            self.data["quantity"] = total_quantity._toDict()
+        if sample_template_id is not None:
+            self.data["templateId"] = sample_template_id
+        if fields is not None:
+            self.data["fields"] = fields
+        ## fail early
+        if attachments is not None:
+            if not isinstance(attachments, list):
+                raise ValueError("attachments must be a list of open files")
+
+
+class TargetLocation:
+    """
+    Base class of target locations. It is recommended to use one of the subclasses
+    and not instantiate this directly.
+    """
+
+    def __init__(self, target_container: Union[str, int, dict, Container]):
+        """
+        Parameters
+        ----------
+        target_container : Union[str, int, dict, Container]
+            A numeric or global ID of a container, or a Container object, or a dict of a Container,
+            or a string: 'w' for workbench, 't' for top-level.
+        Raises
+        ------
+        ValueError
+            If the global id is not 'IC' or the argument is a dict but not that of a container
+        TypeError
+            If type is not supported
+        """
+        self.data = {}
+        if target_container == "t":
+            self.data["removeFromParentContainerRequest"] = True
+        elif target_container == "w":
+            self.data = {}
+
+        elif Id.is_valid_id(target_container):
+            parent_id = Id(target_container)
+            if not parent_id.is_container(True):
+                raise ValueError("Id must be that of a container")
+            self.data["parentContainers"] = [{"id": parent_id.as_id()}]
+        else:
+            raise TypeError("location must be 'w', 't' or a Container, id or global Id")
+
+    def __repr__(self):
+        return f"{self.__class__.__name__}: {self.data!r}"
+
+
+class BenchTargetLocation(TargetLocation):
+    def __init__(self):
+        super().__init__("w")
+
+
+class TopLevelTargetLocation(TargetLocation):
+    def __init__(self):
+        super().__init__("t")
+
+
+class ListContainerTargetLocation(TargetLocation):
+    def __init__(self, target_container: Union[str, int, dict, Container]):
+        super().__init__(target_container)
+
+
+class ImageContainerTargetLocation(TargetLocation):
+    """
+    An location in an ImageContainer is specified by the the container, and the
+    ID of the location within the container.
+    """
+
+    def __init__(
+        self,
+        target_container: Union[str, int, dict, Container],
+        target_location_id: int,
+    ):
+        super().__init__(target_container)
+        self.data["parentLocation"] = {"id": target_location_id}
+        del self.data["parentContainers"]
+
+
+class GridContainerTargetLocation(TargetLocation):
+    """
+    Defines the identity of a grid location to move into, and its coordinates in the grid.
+    """
+
+    def __init__(
+        self,
+        target_container: Union[str, int, dict, Container],
+        col_index: int,
+        row_index: int,
+    ):
+        super().__init__(target_container)
+        gl = GridLocation(col_index, row_index)  ## validates coords
+        self.data["parentLocation"] = {"coordX": gl.x, "coordY": gl.y}
+
+
+class ContainerPost(ItemPost):
+    """
+    Base class for defining a new Container
+    """
+
+    def __init__(
+        self,
+        name: str,
+        tags: Optional[str] = None,
+        description: Optional[str] = None,
+        extra_fields: Optional[Sequence] = [],
+        can_store_containers: bool = True,
+        can_store_samples: bool = True,
+        location: TargetLocation = TopLevelTargetLocation(),
+    ):
+        super().__init__(name, tags, description, extra_fields)
+        if not can_store_containers and not can_store_samples:
+            raise ValueError(
+                "At least one of 'canStoreContainers' and 'canStoreSamples' must be True"
+            )
+        self.data["type"] = "CONTAINER"
+        self.data["canStoreContainers"] = can_store_containers
+        self.data["canStoreSamples"] = can_store_samples
+        self.data.update(location.data)
+
+    def __repr__(self):
+        return f"{self.__class__.__name__}: {self.data!r}"
+
+
+class ListContainerPost(ContainerPost):
+    """
+    Define a new ListContainer to create
+    """
+
+    def __init__(
+        self,
+        name: str,
+        tags: Optional[str] = None,
+        description: Optional[str] = None,
+        extra_fields: Optional[Sequence] = [],
+        can_store_containers: bool = True,
+        can_store_samples: bool = True,
+        location: TargetLocation = TopLevelTargetLocation(),
+    ):
+        super().__init__(
+            name,
+            tags,
+            description,
+            extra_fields,
+            can_store_containers,
+            can_store_samples,
+            location,
+        )
+        self.data["cType"] = "LIST"
+
+
+class ImageContainerPost(ContainerPost):
+    """
+    Define a new ImageContainer to create.
+    """
+
+    def __init__(
+        self,
+        name: str,
+        image_file_path: str,
+        locations: Optional[Sequence] = [],
+        tags: Optional[str] = None,
+        description: Optional[str] = None,
+        extra_fields: Optional[Sequence] = [],
+        can_store_containers: bool = True,
+        can_store_samples: bool = True,
+        location: TargetLocation = TopLevelTargetLocation(),
+    ):
+        """
+        Parameters
+        ----------
+        name : str
+            Name of the image container.
+        image_file_path : str
+            A full file path to the image to use.
+        locations : Optional[Sequence], optional
+            An optional list of (x,y) coordinate tuples of marked locations.
+        tags : Optional[str], optional
+            Comma separated tags
+        description : Optional[str], optional
+        extra_fields : Optional[Sequence], optional
+        can_store_containers : bool, optional
+            The default is True.
+        can_store_samples : bool, optional
+            The default is True.
+        location : TargetLocation, optional
+            The default is TopLevelTargetLocation.
+        """
+        super().__init__(
+            name,
+            tags,
+            description,
+            extra_fields,
+            can_store_containers,
+            can_store_samples,
+            location,
+        )
+        with open(image_file_path, "rb") as img_file:
+            image_b64 = base64.b64encode(img_file.read()).decode("ascii")
+            self.data["newBase64LocationsImage"] = "data:image/png;base64," + str(
+                image_b64
+            )
+        locs = [{"coordX": p[0], "coordY": p[1]} for p in locations]
+        self.data["locations"] = locs
+        self.data["cType"] = "IMAGE"
+
+
+class GridContainerPost(ContainerPost):
+    """
+    Define a new grid container to create
+    """
+
+    def __init__(
+        self,
+        name: str,
+        row_count: int,
+        column_count: int,
+        tags: Optional[str] = None,
+        description: Optional[str] = None,
+        extra_fields: Optional[Sequence] = [],
+        can_store_containers: bool = True,
+        can_store_samples: bool = True,
+        location: TargetLocation = TopLevelTargetLocation(),
+    ):
+        super().__init__(
+            name,
+            tags,
+            description,
+            extra_fields,
+            can_store_containers,
+            can_store_samples,
+            location,
+        )
+        self.data["cType"] = "LIST"
+        self.data["gridLayout"] = {
+            "columnsNumber": column_count,
+            "rowsNumber": row_count,
+        }
+
+
 class InventoryClient(ClientBase):
     """
     Wrapper around RSpace Inventory API.
@@ -685,14 +1059,42 @@ class InventoryClient(ClientBase):
 
         return f"{self.rspace_url}/api/inventory/{self.API_VERSION}"
 
+    MAX_BULK = 100
+    ## Helper method for generic bulk post
+    def _do_bulk(self, post_json):
+        resp_json = self.retrieve_api_results(
+            "/bulk", request_type="POST", params=post_json
+        )
+        return BulkOperationResult(resp_json)
+
+    def bulk_create_sample(self, *sample_posts):
+        """
+        Create up to MAX_BULK samples at once.
+        Parameters
+        ----------
+        *sample_posts : An unpacked iterable of >=1 SamplePost objects
+            Up to MAX_BULK SamplePost objects can be sent at once.
+
+        Returns
+        -------
+        BulkOperationResult
+        """
+        if len(sample_posts) > InventoryClient.MAX_BULK:
+            raise ValueError(
+                f"Max permitted samples is {InventoryClient.MAX_BULK} but was {len(sample_posts)}"
+            )
+        toPost = [s.data for s in sample_posts]
+        bulk_post = {"operationType": "CREATE", "records": toPost}
+        return self._do_bulk(bulk_post)
+
     def create_sample(
         self,
         name: str,
         tags: Optional[str] = None,
         description: Optional[str] = None,
+        extra_fields: Optional[Sequence] = [],
         sample_template_id=None,
         fields=None,
-        extra_fields: Optional[Sequence] = [],
         storage_temperature_min: StorageTemperature = None,
         storage_temperature_max: StorageTemperature = None,
         expiry_date: datetime.datetime = None,
@@ -708,27 +1110,24 @@ class InventoryClient(ClientBase):
         Note that including files to attach to Attachment fields is not supported
         by this method.
         """
-        data = self._set_core_properties(name, tags, description, extra_fields)
-        if storage_temperature_min is not None:
-            data["storageTempMin"] = storage_temperature_min._toDict()
-        if storage_temperature_max is not None:
-            data["storageTempMax"] = storage_temperature_max._toDict()
-        if expiry_date is not None:
-            data["expiryDate"] = expiry_date.isoformat()
-        if subsample_count is not None:
-            data["newSampleSubSamplesCount"] = subsample_count
-        if total_quantity is not None:
-            data["quantity"] = total_quantity._toDict()
-        if sample_template_id is not None:
-            data["templateId"] = sample_template_id
-        if fields is not None:
-            data["fields"] = fields
-        ## fail early
-        if attachments is not None:
-            if not isinstance(attachments, list):
-                raise ValueError("attachments must be a list of open files")
+        toPost = SamplePost(
+            name,
+            tags,
+            description,
+            extra_fields,
+            sample_template_id,
+            fields,
+            storage_temperature_min,
+            storage_temperature_max,
+            expiry_date,
+            subsample_count,
+            total_quantity,
+            attachments,
+        )
 
-        sample = self.retrieve_api_results("/samples", request_type="POST", params=data)
+        sample = self.retrieve_api_results(
+            "/samples", request_type="POST", params=toPost.data
+        )
         if attachments is not None:
             self.serr(f"adding {len(attachments)} attachments")
             for file in attachments:
@@ -870,7 +1269,6 @@ class InventoryClient(ClientBase):
         Returns
         -------
         None.
-
         """
         id_to_delete = Id(sample_id)
         self.doDelete("samples", id_to_delete.as_id())
@@ -937,7 +1335,6 @@ class InventoryClient(ClientBase):
         Returns
         -------
         A list of split subsamples
-
         """
 
         def _do_call(ss_id, params):
@@ -1009,7 +1406,6 @@ class InventoryClient(ClientBase):
         Returns
         -------
         The duplicated item
-
         """
         id_to_copy = Id(item_to_duplicate)
         endpoint = id_to_copy.get_api_endpoint()
@@ -1048,23 +1444,6 @@ class InventoryClient(ClientBase):
             params["resultType"] = result_type.name
         return self.retrieve_api_results("/search", params=params)
 
-    def _set_core_properties(
-        self,
-        name: str,
-        tags: Optional[str] = None,
-        description: Optional[str] = None,
-        extra_fields: Optional[Sequence] = [],
-    ):
-        data = {}
-        data["name"] = name
-        if tags is not None:
-            data["tags"] = tags
-        if description is not None:
-            data["description"] = description
-        if extra_fields is not None:
-            data["extraFields"] = [ef.data for ef in extra_fields]
-        return data
-
     def add_note_to_subsample(
         self, subsample: Union[str, int, dict], note: str
     ) -> dict:
@@ -1090,19 +1469,44 @@ class InventoryClient(ClientBase):
         result = self.retrieve_api_results("/workbenches")
         return [wb for wb in result["containers"]]
 
-    def _configure_parent_container_post(self, location, data):
-        is_wb = False
-        if location == "t":
-            data["removeFromParentContainerRequest"] = True
-        elif location == "w":
-            is_wb = True
-        elif isinstance(location, int) or not is_wb:
-            parent_id = Id(location)
-            if not parent_id.is_container(True):
-                raise ValueError("Id must be that of a container")
-            data["parentContainers"] = [{"id": parent_id.as_id()}]
-        else:
-            raise TypeError("location must be 'w', 't' or a container id or global Id")
+    def bulk_create_container(self, *container_posts):
+        """
+        Create up to MAX_BULK containers at once.
+        Parameters
+        ----------
+        *container_posts : An unpacked iterable of >=1 ContainerPost objects
+            Up to MAX_BULK ContainerPost objects can be sent at once.
+
+        Returns
+        -------
+        BulkOperationResult
+        """
+
+        if len(container_posts) > InventoryClient.MAX_BULK:
+            raise ValueError(
+                f"Max permitted samples is {InventoryClient.MAX_BULK} but was {len(container_posts)}"
+            )
+        toPost = [c.data for c in container_posts]
+        bulk_post = {"operationType": "CREATE", "records": toPost}
+        return self._do_bulk(bulk_post)
+
+    def create_image_container(self, imageContainerPost: ImageContainerPost):
+        """
+        Create a single image container
+
+        Parameters
+        ----------
+        imageContainerPost : ImageContainerPost
+            A definition of the ImageContainerPost to create.
+
+        Returns
+        -------
+        container : A dict of JSON data representing the newly created container.
+        """
+        container = self.retrieve_api_results(
+            "/containers", request_type="POST", params=imageContainerPost.data
+        )
+        return container
 
     def create_list_container(
         self,
@@ -1111,15 +1515,23 @@ class InventoryClient(ClientBase):
         description: Optional[str] = None,
         extra_fields: Optional[Sequence] = [],
         can_store_containers: bool = True,
-        can_store_subsamples: bool = True,
-        location: Union[str, int] = "t",
+        can_store_samples: bool = True,
+        location: TargetLocation = TopLevelTargetLocation(),
     ) -> dict:
+        """
+        Creates a single List Container, either 'top-level',  on the Workbench,
+         or inside an existing Grid or List Container
 
-        data = self._set_core_properties(name, tags, description, extra_fields)
-        data["cType"] = "LIST"
-        data["canStoreContainers"] = can_store_containers
-        data["canStoreSubsamples"] = can_store_subsamples
-        self._configure_parent_container_post(location, data)
+        """
+        data = ListContainerPost(
+            name,
+            tags,
+            description,
+            extra_fields,
+            can_store_containers,
+            can_store_samples,
+            location,
+        ).data
 
         container = self.retrieve_api_results(
             "/containers", request_type="POST", params=data
@@ -1139,24 +1551,68 @@ class InventoryClient(ClientBase):
         description: Optional[str] = None,
         extra_fields: Optional[Sequence] = [],
         can_store_containers: bool = True,
-        can_store_subsamples: bool = True,
-        location: Union[int, str] = "t",
+        can_store_samples: bool = True,
+        location: TargetLocation = TopLevelTargetLocation(),
     ) -> dict:
+        """
+        Parameters
+        ----------
+        name : str
+            The container name.
+        row_count : int
+            Then number of rows (max 24).
+        column_count : int
+            The number of columns (max 24)
+        tags : Optional[str], optional
+            Comma-separated
+        description : Optional[str], optional
+            A description for this contaier. The default is None.
+        extra_fields : Optional[Sequence], optional
+            One or more ExtraFields. The default is [].
+        can_store_containers : bool, optional
+            Whether this container can store containers inside it. The default is True.
+        can_store_samples : bool, optional
+            Whether this container can store subsamples inside it. The default is True.
+        location : TargetLocation
+            A subclass of TargetLocation. Defaults to TopLevel
+        Returns
+        -------
+        dict
+            The created container.
+        """
 
-        data = self._set_core_properties(name, tags, description, extra_fields)
+        data = GridContainerPost(
+            name,
+            row_count,
+            column_count,
+            tags,
+            description,
+            extra_fields,
+            can_store_containers,
+            can_store_samples,
+            location,
+        ).data
         data["cType"] = "GRID"
-        data["canStoreContainers"] = can_store_containers
-        data["canStoreSubsamples"] = can_store_subsamples
-        data["gridLayout"] = {"columnsNumber": column_count, "rowsNumber": row_count}
-
-        self._configure_parent_container_post(location, data)
 
         container = self.retrieve_api_results(
             "/containers", request_type="POST", params=data
         )
         return container
 
-    def set_as_top_level_container(self, container: Union[int, str, dict, Container]):
+    def set_as_top_level_container(
+        self, container: Union[int, str, dict, Container]
+    ) -> dict:
+        """
+        Moves a container from its current location to be a top-level container
+
+        Parameters
+        ----------
+        container : Union[int, str, dict, Container]
+            Id, dict or container object.
+        Returns
+        -------
+        The updated container
+        """
         data = {"removeFromParentContainerRequest": True}
         c_id = Id(container)
 
@@ -1164,9 +1620,15 @@ class InventoryClient(ClientBase):
             f"/containers/{c_id.as_id()}", request_type="PUT", params=data
         )
 
+    def _id_as_container_id(self, target_container_id):
+        id_target = Id(target_container_id)
+        if not id_target.is_container(maybe=True):
+            raise ValueError("Target must be a container")
+        return id_target
+
     def add_items_to_list_container(
         self,
-        target_container_id: Union[str, int],
+        target_container_id,
         *item_ids: str,
     ) -> list:
         """
@@ -1188,12 +1650,9 @@ class InventoryClient(ClientBase):
         Returns
         -------
         BulkoperationResult
-
         """
-        id_target = Id(target_container_id)
-        if not id_target.is_container(maybe=True):
-            raise ValueError("Target must be a container")
 
+        id_target = self._id_as_container_id(target_container_id)
         valid_item_ids = []
 
         ## assert there are no invalid globai ids
@@ -1207,6 +1666,43 @@ class InventoryClient(ClientBase):
 
         return self._do_add_to_list_container(valid_item_ids, id_target)
 
+    def add_items_to_image_container(
+        self,
+        target_container_id: Union[str, int, dict],
+        items_to_move: Sequence,
+        location_ids: Sequence,
+    ) -> BulkOperationResult:
+        """
+        Adds a list of items to move to a list of  locations in an image container.
+        The 2 lists must be of equal length. If unequal length, items in the longer list
+         will be ignored once items in the shorter list are exhausted (like 'zip' function)
+        Parameters
+        ----------
+        target_container_id : Union[str, int, dict]
+            An identifier for the container.
+        items_to_move : Sequence
+            A list of globalIds or dicts of items to move.
+        location_ids : Sequence
+            A List of location identifiers that are the ids of image container locations.
+        Returns
+        -------
+        BulkOperationResult
+        """
+        self._id_as_container_id(target_container_id)
+
+        loci = []
+        for (item, loc_id) in zip(items_to_move, location_ids):
+            item_id = Id(item)
+            loci.append(
+                {
+                    "type": item_id.get_type(),
+                    "id": item_id.as_id(),
+                    "parentLocation": {"id": loc_id},
+                }
+            )
+        bulk_post = {"operationType": "MOVE", "records": loci}
+        return self._do_bulk(bulk_post)
+
     def add_items_to_grid_container(
         self,
         target_container_id: Union[str, int, GridContainer],
@@ -1225,32 +1721,21 @@ class InventoryClient(ClientBase):
         ------
         ValueError
             If items are the wrong or inconsistent type
-
         Returns
         -------
         list
             A list of updated items showing their current position
-
         """
         if isinstance(target_container_id, GridContainer):
             if target_container_id.free() < len(grid_placement.items_to_move):
                 raise ValueError(
                     f"not enough space in {target_container_id.data['globalId']} to store {len(grid_placement.items_to_move)} - only {target_container_id.free()} spaces free."
                 )
-        id_target = Id(target_container_id)
-        if not id_target.is_container(maybe=True):
-            raise ValueError("Target must be a container")
+        id_target = self._id_as_container_id(target_container_id)
         ## assert there are no invalid global ids (things that are not subsamples)
 
         bulk_post = self._create_bulk_move(id_target, grid_placement)
-        ## get target - are there enough spaces?
-        ## iterate over grid (0 or 1 based?)
-        ## use bulk API?
-
-        resp_json = self.retrieve_api_results(
-            "/bulk", request_type="POST", params=bulk_post
-        )
-        return BulkOperationResult(resp_json)
+        return self._do_bulk(bulk_post)
 
     def _do_add_to_list_container(self, items, id_target):
         coords = []
@@ -1263,12 +1748,7 @@ class InventoryClient(ClientBase):
                 }
             )
         to_post = {"operationType": "MOVE", "records": coords}
-
-        resp_json = self.retrieve_api_results(
-            "/bulk", request_type="POST", params=to_post
-        )
-
-        return BulkOperationResult(resp_json)
+        return self._do_bulk(to_post)
 
     def _create_bulk_move(self, grid_id: Id, gp: GridPlacement):
         coords = []  # array of x,y coords
@@ -1420,7 +1900,6 @@ class InventoryClient(ClientBase):
         -------
         Dict
             The newly created template.
-
         """
         return self.retrieve_api_results(
             "/sampleTemplates", request_type="POST", params=sample_template_post
@@ -1497,7 +1976,6 @@ class InventoryClient(ClientBase):
         Returns
         -------
         void, no return value
-
         """
         st_id = Id(sample_template_id)
         url_base = self._get_api_url()
@@ -1539,7 +2017,6 @@ class InventoryClient(ClientBase):
         -------
         dict
             The updated template.
-
         """
         id_to_restore = Id(sample_template_id)
         return self.retrieve_api_results(
@@ -1563,7 +2040,6 @@ class InventoryClient(ClientBase):
         -------
         dict
             The updated sample template.
-
         """
         st_id = Id(sample_template_id)
         return self._do_transfer_owner("sampleTemplates", st_id, new_owner)
@@ -1582,7 +2058,6 @@ class InventoryClient(ClientBase):
         -------
         dict
             The updated sample.
-
         """
         sample_id = Id(sample_id)
         return self._do_transfer_owner("samples", sample_id, new_owner)
